"""
PYPOWER version.
"""


def ppver(*args):
    """
    Returns PYPOWER version info for current installation.

    @author: Ray Zimmerman (PSERC Cornell)
    """

    ver = {'Name': 'PYPOWER',
           'Version': '5.1.4',
           'Release':  '',
           'Date': '27-June-2018'}

    return ver
